@@ -19,6 +19,9 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <config.h>
+
+#include <glib/gi18n-lib.h>
 #include "util.h"
 #include "xutils.h"
 #include "private.h"
@@ -259,6 +262,52 @@ _wnck_get_client_type ()
   return client_type;
 }
 
+/**
+ * _make_gtk_label_bold.
+ *
+ * Switches the font of label to a bold equivalent.
+ * @label: The label.
+ **/
+void
+_make_gtk_label_bold (GtkLabel *label)
+{
+  PangoFontDescription *font_desc;
+
+  font_desc = pango_font_description_new ();
+
+  pango_font_description_set_weight (font_desc,
+                                     PANGO_WEIGHT_BOLD);
+
+  /* This will only affect the weight of the font, the rest is
+   * from the current state of the widget, which comes from the
+   * theme or user prefs, since the font desc only has the
+   * weight flag turned on.
+   */
+  gtk_widget_modify_font (GTK_WIDGET (label), font_desc);
+
+  pango_font_description_free (font_desc);
+}
+
+void
+_make_gtk_label_normal (GtkLabel *label)
+{
+  PangoFontDescription *font_desc;
+
+  font_desc = pango_font_description_new ();
+
+  pango_font_description_set_weight (font_desc,
+                                     PANGO_WEIGHT_NORMAL);
+
+  /* This will only affect the weight of the font, the rest is
+   * from the current state of the widget, which comes from the
+   * theme or user prefs, since the font desc only has the
+   * weight flag turned on.
+   */
+  gtk_widget_modify_font (GTK_WIDGET (label), font_desc);
+
+  pango_font_description_free (font_desc);
+}
+
 void
 _wnck_init (void)
 {
